#ifndef CGAL_LIN_APPROXIMANT_2_H
#define CGAL_LIN_APPROXIMANT_2_H

#include "Curve_constants.h"
#include "Helpers.h"
#include <CGAL/squared_distance_2.h>
#include <CGAL/intersections.h>
#include <CGAL/geowin_support.h>


// Template class Piecewise_lin_approximant_2 implements a piecewise linear
// approximant for a curve. This is a vector of pairs (point, paramter_value)
// at which the piecewise linear approximant interpolates the curve. The 
// points are linked with line segments.
// Template parameters:
//    NumType_   number type which is used to store coordinates
//
//    PointIt    InputIterator to a container of points, by default 
//          vector<CGAL::Point_2<CGAL::Cartesian<NumType_> > >::const_iterator
//
//    ParamIt    InputIterator to a container of NumType_ values (the 
//             parameter values), by default vector<NumType_>::const_iterator>

template <class NumType_, 
    class PointIt = vector<CGAL::Point_2<CGAL::Cartesian<NumType_> > >::const_iterator, 
    class ParamIt = vector<NumType_>::const_iterator >
class Piecewise_lin_approximant_2 {

public:
    typedef NumType_ NumType;
    typedef vector<NumType> ParamVector;
    typedef ParamVector::const_iterator ParamVectorIt;
    typedef CGAL::Cartesian<NumType> REP;
    typedef CGAL::Point_2<REP> Point;
    typedef vector<Point> Points;
    typedef Points::const_iterator PointVectorIt;
    typedef std::list<Point> PointList;    
    typedef CGAL::Segment_2<REP> Segment;
    typedef list<Segment> SegmentList;    
    typedef vector<Segment> Segments; 
    //typedef CGAL::Bbox_2 Bbox;       
    typedef CGAL::Vector_2<REP> Vector;
    typedef CGAL::Object Object; 
       

    // Piecewise_lin_approximant_2: void -> Piecewise_lin_approximant_2    
    // Effect: contructs a empty Piecewise_lin_approximant_2 object 
    // parameterless contructor
    Piecewise_lin_approximant_2(): points(), number(0), t_values() {	
    }    

    // Piecewise_lin_approximant_2: PointIt, PointIt, NumType, NumType 
    //                                   -> Piecewise_lin_approximant_2
    // Parameter:
    //    first     iterator to container of points
    //    last      iterator to container of points
    // Pre:
    //     [first, last) is a valid range
    // Effect: contructs a Piecewise_lin_approximant_2 object with uniform
    // parametrisation over [a,b] and points the points in 
    // the range [first, last)    
    Piecewise_lin_approximant_2(const PointIt first, const PointIt last, 
				NumType a=LOW_BOUND, NumType b=UP_BOUND) {
	number = distance(first, last);
	points.resize(number);	
	copy(first, last, points.begin());

	uniform_parametrisation(t_values, a, b, number);	
    }  

    // Piecewise_lin_approximant_2: PointIt, PointIt, ParamIt ->
    //                              Piecewise_lin_approximant_2
    // Parameters:
    //    first_pt, last_pt     iterators to the container with the points
    //                          of this linear approximant    
    //    first_t               iterator to the container with the parameter 
    //                          values at wich the points of this linear 
    //                          approximat are calculated    
    // Pre:
    //    [first_pt, last_pt)                          is a valid range
    //    [first_t, first_t + (last_pt - first_pt))    is a valid range
    // Effect: contructs a Piecewise_lin_approximant_2 object from a point 
    // container and a parameter value container
    Piecewise_lin_approximant_2(const PointIt first_pt, const PointIt last_pt,
				const ParamIt first_t) {        
	set_points(first_pt, last_pt, first_t);	
    } 

    // copy constructor: Piecewise_lin_approximant_2 -> 
    //                                           Piecewise_lin_approximant_2 
    Piecewise_lin_approximant_2(const Piecewise_lin_approximant_2& c) {	
	set_points(c.points_begin(), c.points_end(), 
		   c.parametrisation_begin());
    }

    virtual ~Piecewise_lin_approximant_2() {}    
    
    // set_points: PointIt, PointIt, ParamIt -> void
    // Parameters:
    //    first_pt, last_pt     iterators to the container with the new points
    //                          for this linear approximant    
    //    first_t               iterators to the container with the new 
    //                          parameter values for this approximat 
    // Pre:
    //    [first_pt, last_pt)                         is a valid range
    //    [first_t, first_t + (last_pt - first_pt))   is a valid range
    // Effect: changes the points of the linear approximant and updates
    //         number and t_values
    void set_points(const PointIt first_pt, const PointIt last_pt, 
		    const ParamIt first_t) {   
	number = distance(first_pt, last_pt);
	points.resize(number);	
	copy(first_pt, last_pt, points.begin());	
	t_values.resize(number);
	copy_n(first_t, number, t_values.begin());	 
    }    

    // get_number: void -> unsigned
    // Returns: the number of points in the linear approximant
    unsigned get_number() const {
	return number;
    }

    // points_begin(): void -> PointVectorIt
    // Returns: const_iterator to the first point of this linear approximant
    PointVectorIt points_begin() const {	
	return points.begin();
    }

    // points_end(): void -> PointVectorIt
    // Returns: the past-the-end iterator to the point vector
    //          of this linear approximant
    PointVectorIt points_end() const {	
	return points.end();
    }

    // parametrisation_begin(): void -> ParamVectorIt
    // Returns: const_iterator to the first parameter value of 
    //          this linear approximant
    ParamVectorIt parametrisation_begin() const {
	return t_values.begin();
    }

    // parametrisation_end(): void -> ParamVectorIt
    // Returns: past-the-end iterator to the container with parameter values of    //          this linear approximant
    ParamVectorIt parametrisation_end() const {
	return t_values.end();
    }    

    // evaluate_at: NumType -> Point
    // Parameters:
    //    t     a value in [t_values[0], t_values[number-1]]
    // Return:  a point on the linear approximant at t
    // Pre: t_values[0] <= t <= t_values[number-1]
    Point evaluate_at(NumType t) {
	assert((t_values[0]<=t) && (t <= t_values[number-1]));	
	unsigned i=0;
	while((i<number) && (t>=t_values[i])) i++;
	if(i==number) return points[number-1];	
	NumType coor_low = (t_values[i]-t)/(t_values[i]-t_values[i-1]);
	NumType coor_up = (t-t_values[i-1])/(t_values[i]-t_values[i-1]);
	Vector tmp = (points[i-1]-CGAL::ORIGIN)*coor_low + 
	    (points[i]-CGAL::ORIGIN)*coor_up;     
	return (CGAL::ORIGIN + tmp);
    }
        
    // draw: void -> void       
    // Effect: opens a window of a GeoWin object and draws the piecewise
    //         linear approximant
    void draw() const {       
	GeoWin GW("Piecewise Linear Approximant Demo");	
	geowin_init_default_type((SegmentList*)0, leda_string("SegmentList"));

	SegmentList segments;
	PointList pts_l(points.begin(), points.end());
	link_points_to_segments(pts_l, segments);	
	geo_scene a_scene = GW.new_scene(segments);
	GW.set_color(a_scene, leda_blue);
	GW.set_active_line_width(a_scene, 3);
	GW.set_visible(a_scene, true);
	
	GW.display(80, 10);	
	Bbox b = bbox();
	GW.init(b.xmin(), b.xmax(), b.ymin(), b.ymax(), 5);	
	GW.edit(a_scene);	
    };
    
    // min_squared_distance: Point -> NumType
    // Parameters:
    //      pt the point to which the minimal squared distance is computed
    // Returns: the minimal squared distance from pt to the linear approximant
    // Pre: the linear approximant has at least one point, i.e.points.size()>0
    NumType min_squared_distance(const Point& pt) const {	
	assert(number>0);
	if(number<2) return squared_distance(points[0], pt);
	NumType squared_dist;
	Segment tmp(points[0], points[1]);
	NumType min = squared_distance(tmp, pt);
	for(unsigned i=2; i<number; i++) {
	    Segment s(points[i-1], points[i]);	   
	    squared_dist = squared_distance(s, pt);	    
	    if(squared_dist==0) return squared_dist;
	    if(squared_dist < min) min=squared_dist;
	}
	return min;
    }

    // bbox: NumType -> Bbox
    // Parameters:     
    //     eps     the points which lies at distance <= eps from the linear 
    //             approximant assumes to be in the bounding box, 
    //             default value for eps 0
    // Returns: a bounding box of this piecewise linear approximant 
    //          containing all the points at a distance <= eps
    // Pre: number >= 1    
    Bbox bbox(NumType epsilon=0) const {	
	assert(number>=1);		
	double x_min, x_max, y_min, y_max, tmp;	
	x_min=x_max=_to_float(points[0].x());		
	y_min=y_max=_to_float(points[0].y());	
	for(unsigned i=1; i<number; i++) {
	    tmp = _to_float(points[i].x());
	    if(tmp<x_min) x_min = tmp;
	    else if(tmp>x_max) x_max = tmp;
	    tmp = _to_float(points[i].y());
	    if(tmp<y_min) y_min = tmp;
	    else if(tmp>y_max) y_max = tmp;   
	}	
	tmp = _to_float(epsilon);
	Bbox box(x_min-tmp, y_min-tmp, x_max+tmp, y_max+tmp);
	return box;
    }

    // has_on: Point, NumType -> bool
    // Parameters: 
    //     pt        the point is tested if it is on the piecewise 
    //               linear approximant
    //     epsilon   if dist(pt, piecewise linear approximant)<=epsilon
    //               we assume pt lies on it, default value EPSILON
    // Returns: true, if distance(pt, piecewise linear approximant)<=epsilon
    //          false otherwise    
    bool has_on(const Point& pt, NumType epsilon=EPSILON) const {	  
        if(number==0) return false;
	Bbox pt_bbox = pt.bbox();		
	if(!do_overlap(pt_bbox, bbox(epsilon))) {	    
	    return false;	    
	}
	else {	    
	    if(number==1) {		
		if(squared_distance(points[0], pt)<=(epsilon*epsilon)) 
		     return true;
		else return false;
	    }	    
	    else if(number==2) {		
		Segment s(points[0], points[1]);
		if(squared_distance(s, pt)<=(epsilon*epsilon)) return true;
		else return false;
	    }
	    else {			
		Piecewise_lin_approximant_2 left, right;		    
		divide(left, right);
		bool left_res = left.has_on(pt);
		if(left_res) {		    
		    return true;
		}
		else {		    
		    return right.has_on(pt);		    
		}		
	    }
	}
    }

    // intersections: Piecewise_lin_approximant_2 -> Points
    // Parameters:
    //    approx        another piecewise linear approximant
    //    intersec_pts  the vector of intersections points of the two
    //                  linear approximants    
    //  intersec_pts is empty at the begining of intersections
    void intersections(Points & intersec_pts, 
		       const Piecewise_lin_approximant_2 & approx,
		       NumType epsilon=EPSILON) const {
	//one of the approximants is empty
	if(number==0 || approx.get_number()==0) return;
	//determine which one is the shorter
	Piecewise_lin_approximant_2 shorter, longer;
	if(number<=approx.get_number()) {
	    shorter = (*this);
	    longer = approx;
	}
	else {
	    shorter = approx;
	    longer = (*this);
	}
	//if bboxes do not overlap return
        if(!do_overlap(shorter.bbox(), longer.bbox())) return;
	
	PointVectorIt it_s, it_l;
	it_s = shorter.points_begin();	
	it_l = longer.points_begin();
	//if the shorter approximant has just 1 point call has_on 
	//on the longer approximant with this point
        if(shorter.get_number()==1) {
	    Point s_pt = (* it_s);
	    if(longer.has_on(s_pt, epsilon)) {
		//omiting to insert duplicates
		if(intersec_pts.empty() || 
		   intersec_pts[intersec_pts.size()-1]!=s_pt)
		     intersec_pts.push_back(s_pt);
	    }
	    return;
        }	
	//both approximants are segments
	if((shorter.get_number()==2) && (longer.get_number()==2)) {	    
	    Segment s_s(*it_s, *(it_s+1));    
	    Segment s_l(*it_l, *(it_l+1));
	    //check if they intersects		
	    Point i_pt;
	    Segment i_segm;
	    Object i_result = intersection(s_s, s_l);
	    if(assign(i_pt,i_result)) {
		if(intersec_pts.empty() || 
		   intersec_pts[intersec_pts.size()-1]!=i_pt)
		     intersec_pts.push_back(i_pt);		    
	    }
	    else if(assign(i_segm,i_result)) {		
		//omiting to insert duplicates
		if(intersec_pts.empty() || 
		   intersec_pts[intersec_pts.size()-1]!=i_segm.source())
		      intersec_pts.push_back(i_segm.source());
		intersec_pts.push_back(i_segm.target());
	    }
	    return;
	} //end both are segmnets	    	
	//at least the longer  approximant has > 2 points
	//divide the longer  approximant, and test the two halfs 
	//for inresections with the shorter approximant
	Piecewise_lin_approximant_2 f,s;
	longer.divide(f,s);
	//test the first and the second half for intersections
	shorter.intersections(intersec_pts, f);
	shorter.intersections(intersec_pts, s);
	return;	
    }

    // divide: Piecewise_lin_approximant_2 ->
    //               Piecewise_lin_approximant_2, Piecewise_lin_approximant_2
    // Parameters:
    //   first    a piecewise linear approximant defined by the first half 
    //            of the points
    //   second   a piecewise linear approximant defined by the other half of 
    //            the points
    // Pre: number > 2
    // Effect: divides this piecewise linear approximant in two new objects
    //         first containing the fisrt number/2 (point, paramand) pairs
    //         and second containing the remaining pairs 
    void divide(Piecewise_lin_approximant_2 & first, 
		Piecewise_lin_approximant_2 & second) const {
	Points first_v, second_v;	
	ParamVector first_t, second_t;
	if(number<=2) return;
	
	unsigned half = number/2;	
	unsigned num_f;
	if((number%2)!=0) {
	   num_f = half + 1; 	
	}
	else num_f = half;
	first_v.resize(num_f);
	copy_n(points.begin(), num_f, first_v.begin());
	first_t.resize(num_f);
	copy_n(t_values.begin(), num_f, first_t.begin());		
	second_v.resize(number-num_f+1);	
	copy_n((points.begin()+num_f-1), (number-num_f+1), second_v.begin());
	second_t.resize(number-num_f+1);
	copy_n((t_values.begin()+num_f-1), (number-num_f+1), 
	       second_t.begin());	
	first.set_points(first_v.begin(), first_v.end(), 
			 first_t.begin());	
	second.set_points(second_v.begin(), second_v.end(), 
			  second_t.begin());
    }

    // discrete_derivative: unsigned, NumType -> 
    //                                  Piecewise_lin_approximant_2, bool
    // Papameters:    
    //    r          compute the r-th discrete derivative of this piecewise 
    //               linear approximant, default value r=1
    //    deriv_approx   contains the r-th discrete derivative, if it exists  
    // Returns: true, if the r-th discrete derivative exists (r<number)
    //          false otherwise       
    bool discrete_derivative(Piecewise_lin_approximant_2 & deriv_approx, 
			     unsigned r) const {	
	if(r>=number) return false;	
	if(r==number-1) return false;	
	Points pts = points;	
	unsigned n = number-1;
	unsigned d = r;
	NumType_ coef;
	Point mid_pt, pt;
	while(d>0) {	   
	    for(unsigned i=0; i<n ; i++) {
		assert(t_values[i+1]!=t_values[i]);		
		//middle = (t_values[i] + t_values[i+1])/(NumType) 2.0;
		mid_pt = midpoint(pts[i], pts[i+1]);
		coef = ((NumType) 2.0)/(t_values[i+1]-t_values[i]);
		if(i==0) pts[i] = CGAL::ORIGIN + (mid_pt - pts[i])*coef; 
		else pts[i] = midpoint(pt, CGAL::ORIGIN + 
					    (mid_pt - pts[i])*coef);
		pt = CGAL::ORIGIN + (pts[i+1] - mid_pt)*coef;		
	    }
	    pts[n] = pt;
	    d--;    
	}	
	deriv_approx.set_points(pts.begin(), pts.end(), 
				t_values.begin());
	return true;
    }

    bool operator==(const Piecewise_lin_approximant_2& other) const{
	if(number==other.get_number()) {	    
	    return (equal(points.begin(), points.end(),other.points_begin()) 
		    &&
		    equal(t_values.begin(), t_values.end(), 
			  other.parametrisation_begin()));
	}
	else return false;
    }

private:
    // the points at which the piecewise linear approximant interpolates the 
    // curve
    Points points;

    // the number of points in this piecewise linear approximant 
    // (points.size())
    unsigned number;

    // a vector with parameter values at wich the points of the 
    // approximant was calculated p(t_values[i])=points[i]
    // t_values[0]=a, t_values[number-1]=b
    ParamVector t_values;
};
#endif //CGAL_LIN_APPROXIMANT_2_H
